const char* vorosketchInfo = 

/**********************************************************************/

"Vorosketch v0.14 BETA by Herman Haverkort, 25 May 2022";

/***********************************************************************

Copyright 2022 Herman Johannes Haverkort

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

New versions of vorosketch.cpp may be available from 
http://herman.haverkort.net.


************************************************************************

WHAT IS VOROSKETCH?

Vorosketch is a simple tool to sketch small Voronoi diagrams as a bitmap. 
Use it wisely. It is not designed for speed or for accuracy: it may miss 
features of the diagram that are narrower than a pixel. But it is easy 
to adapt to different distance measures regardless of whether we 
understand the geometry of the bisectors yet.


************************************************************************

SUPPORTED DISTANCE MEASURES

Below are the supported definitions for d(p,q), where p is any point in 
the plane, and q is a site. To select the distance measure named xxx,
write -m xxx on the command line. One may also select weighted 
combinations, e.g. -m a*xxx+b*yyy combines xxx and yyy with coefficients
a and b.

For point or polyline sites:

  euclidean   minimum euclidean distance from p to any point of q

  squared     square of the above

  manhattan   minimum L1-distance from p to any point of q

For point sites only:

  Lx          (where x is a real number other than zero) minimum 
              Lx-distance from p to q

  min         minimum of distance in x and distance in y (L-inf)

  bbox        product of distance in x and distance in y

  max         maximum of distance in x and distance in y (Linf)

  highway     minimum travel time from p to q when higher speed of
              travel is possible on designated line segments

  karlsruhe   minimum distance from p to q when travelling only along
              rays from the origin and circles centred on the origin

  city        minimum travel time from p to q when speed is linearly
              proportional to the distance from the centre

  azimuth     difference between the azimuth coordinates modulo 2pi with
              respect to the origin

  logradius   logarithm of distance from the origin

  koeln       shorthand for azimuth+logradius: minimum travel time from 
              p to q when travelling only along rays from the origin and 
              circles centred on the origin, while speed is linearly 
              proportional to the distance from the centre

  orbitout    1/|q| - 2/(|q|+|p-q|+|p|)
              this is, modulo constant factors, the minimum kinetic
              energy which we would have to give an object at q in order 
              to reach p, subject to gravitation towards the origin
              (Why? The speed at q in orbit around the origin is 
              proportional to the square root of 1/|q| - 1/a, where a 
              is the major axis of the elliptical orbit. Since p has to 
              be on that orbit, too, we get that 2a is the length of the
              path opfqo, where o is the origin and f is the other focal
              point of the elliptical orbit. To minimise 2a, and thus, 
              the necessary speed at q, we put f on the segment pq.)

  orbitin     1/|p| - 2/(|p|+|q-p|+|q|)
              this is, modulo constant factors, the minimum kinetic
              energy which we would have to give an object at p in order 
              to reach q, subject to gravitation towards the origin
              (note that for any given p, the site q that minimises this
              is simply the site that minimises |q-p| with an additive
              weight of |q|)

For polyline sites only:

  angle       1 / smallest opening angle of any wedge with apex at p 
              that includes q
              this is effectively: how small q looks as seen from p

  detour      the distance to a line segment q1q2 is |q1p|+|pq2|-|q1q2|

  dilation    the distance to a line segment q1q2 is 
              (|q1p|+|pq2|-|q1q2|)/|q1q2|

For ray sites (points q with direction towards a point q'):

  turn        minimum angle between qq' and qp

  leftturn    minimum counterclockwise angle from qq' to qp

  dubins      minimum length of a path from q to p that starts in 
              direction q' and has curvature radius at most |qq'|

To include additional user-defined distance measures from 
userdistances.cpp, compile with the -D INCLUDE_USER_DISTANCES option.


***********************************************************************/
const char* vorosketchFileFormat =

"INPUT AND OUTPUT FORMAT                                                \n\
                                                                        \n\
The sites are read from a text file on standard input that contains a   \n\
sequence of numbers separated by spaces or line breaks.                 \n\
                                                                        \n\
The first number is the number of sites (at most 128). Then there is a  \n\
list of numbers for each site: first the number of points (1 for point  \n\
sites, 2 for ray sites, 2 or more for polyline sites); then for each    \n\
point the x and y coordinate; and finally the weight of the site (for   \n\
unweighted Voronoi Diagrams, just set the weight to 0).                 \n\
                                                                        \n\
If the highway distance is used, the highways must be specified after   \n\
the sites in the same format (with weights interpreted as speed)        \n\
                                                                        \n\
The output is a bitmap in .bmp-format on standard output, that shows    \n\
the Voronoi diagram clipped to the square [-1,-1] x [1,1].              \n";


/***********************************************************************

WEIGHTED DISTANCES

By default, the specified site weights are subtracted from the 
calculated distances. If the command line includes the -d option, then
the distances are divided by the specified weights instead. Thus, we 
can create, among others, the following types of Voronoi diagrams:

additively weighted:         euclidean
multiplicatively weighted:   euclidean -d
power diagram:               squared

Note that definitions in the literature may differ with respect to 
whether weights are added or subtracted, whether they are multiplied or 
divided by, and whether they are squared or not. So be sure to input 
the weights in the file in such a way that you get the intended 
result.


************************************************************************

WHAT IF THERE ARE REGIONS OF NON-ZERO AREA, IN WHICH MULTIPLE SITES ARE 
AT THE SAME DISTANCE?

The *two* sites that are listed in the input file first win. For this 
purpose, the sites have to be at exactly the same calculated distance. 
To account for rounding errors, use the -e c option to regard 
distances as equal if they do not differ more than c times a pixel's
width.

In case of polyline sites, rather than identifying regions that are 
closest to an endpoint shared by two line segments, one may want to 
divide those regions by the angular bisector of those segments. This 
can be achieved with the -s c options, which shortens the first and the
last segment of each site by c times a pixel's width.
 

***********************************************************************/
const char* vorosketchOptions = 

"OPTIONS                                                                \n\
                                                                        \n\
-m xxx   use distance measure xxx. The following measures are supported:\n\
         euclidean (default), squared, manhattan, Lx (for any real      \n\
         number x other than 0), min, bbox, max, karlsruhe, city,       \n\
         azimuth, logradius, koeln, highway, orbitout, orbitin, angle,  \n\
         detour, dilation, turn, leftturn, dubins.                      \n\
         xxx+yyy selects the sum of distance measures xxx and yyy.      \n\
         a*xxx+b*yyy selects a weighted sum with coefficients a and b.  \n\
-d       divide by weights instead of subtracting them                  \n\
-e c     set distance equality threshold to c (default: 0)              \n\
-s c     shorten polyline sites by c times a pixel's width (default: 0) \n\
-2       render 2nd-order Voronoi diagram (cannot be combined with -e)  \n\
-f       render farthest-site Voronoi diagram                           \n\
-a       anonymous: do not draw the sites                               \n\
-b       draw region boundaries in black (default when -c not selected) \n\
-c       draw coloured sites and regions, using 19 colours from Sasha   \n\
         Trubetskoy's list of 20 simple, distinct colours               \n\
-u       draw distance contour line at standard distance from each site \n\
         (0 for subtractively, 1 for divisively weighted diagrams)      \n\
         (note: areas within standard distance from the associated site \n\
         are shaded in any case, unless -2 or -f is selected)           \n\
-i x     draw contour lines at intervals of x units                     \n\
-g x     shading by distance, 50% brightness at distance x              \n\
-r n     set output resolution to n x n pixels (default: 2000)          \n\
-+       mark the origin                                                \n\
-v       verbose: show option settings and progress indicator on        \n\
         standard error output                                          \n\
-?       show how to use this tool                                      \n";


/***********************************************************************

HEADERS


**********************************************************
* Tunable constants etc.                                 *
*********************************************************/

typedef long double DOUBLE;

const int nrPens = 4;

// stroke widths in pixels; must be even, because strokes 
// are centered on pixel corners, not pixel centres:
const int strokeWidth[nrPens] = { 2, 4, 8, 12 };

const int contourPen = 0;
const int bisectorPen = 1;
const int sitePen = 2;
const int hatchPen = 3;
const int siteOutlinePen = 3;

// maximum number of sites (arbitrary); this only to 
// catch mistakes in the input that would cause the 
// program to take forever 
const int MAXNRSITES = 1000;
const int MAXNRHIGHWAYS = 100;

// uncomment the following line for checkerboard instead
// of hatching in shared regions:
#define CHECKERBOARD

// length of progress indicator:
const int progressUnits = 40;


/*********************************************************
* Includes                                               *
*********************************************************/

#include <string>
using std::string;
#include <string.h>
#include <sstream>
using std::stringstream;
#include <iostream>
using std::istream;
using std::ostream;
using std::cin;
using std::cout;
using std::cerr;
using std::endl;
#include <cstdlib>
#include <vector>
#include <queue>
#include <algorithm>
using std::swap;
using std::min;
using std::max;
#include <limits>
const DOUBLE INF = std::numeric_limits<DOUBLE>::max();


/*********************************************************
* Input/output headers                                   *
*********************************************************/

enum ErrorLevel 
{ 
  warning, 
  userError, 
  usageError, 
  internalError 
};

class Complaint 
// an object to which an error message can be witten with <<. 
// when the object is destroyed, the error message is produced
// and, if it is not a mere warning, the program halts.
{
    private:
  static const char* errorLabel[];
  stringstream complaint;
  ErrorLevel errorLevel;
    public:
  Complaint(ErrorLevel _errorLevel = userError);
  Complaint(ErrorLevel _errorLevel, const string& expected, const string& read);
  ~Complaint();
  template <typename T> Complaint& operator<<(const T& t);
};

void makeLowerCase(string& s);

// convert a string to a number, verifying that the string 
// does not contain any further content:
DOUBLE strtofstop(const string& str);
int strtoistop(const string& str);


void reportProgress(DOUBLE x, string description);


class RGB
{
    private:
  int mRed;
  int mGreen;
  int mBlue;
    public:
  RGB(const int _red, const int _green, const int _blue);
  RGB(const double _grey = 1);
  const int red() const;
  const int green() const;
  const int blue() const;
  RGB& operator*=(const RGB& colour);
  static RGB black;
  static RGB grey;
  static RGB white; 
};


class BitmapFile {
    private:
  ostream& outputStream;
  const int width;
  const int padding;
  int samplesYetToCome;
  void writeInteger(int i, const int bytes) { 
    outputStream.write((char*) &i, bytes); 
  }
    public:
  BitmapFile(ostream& _outputStream, const int _width, const int height);
  BitmapFile& writePixel(const RGB& colour);
  ~BitmapFile();
};


/*********************************************************
* Options headers                                        *
*********************************************************/

struct OptionSyntax 
{
  int rank;
  string id;
  bool requiresArgument;
  bool multipleAllowed;
  OptionSyntax(
    int _rank, 
    string _id, 
    bool _requiresValue = false,
    bool _multipleAllowed = false
  );
};


class Distance;


class Options
{
    public:
  Distance* distance;
  bool useHighways;
  enum { subtractive, divisive } weighting;
  DOUBLE equalityThreshold;
  DOUBLE segmentShortening;
  enum { first, second, farthest } order;  
  bool drawSites;
  bool drawRegionBoundaries;
  int nrColours;
  RGB* colours;
  bool drawStandardContours;
  DOUBLE contourInterval;
  bool shadeInnerAreas;
  DOUBLE contourDistance;
  DOUBLE halfBrightDistance;
  int width;
  bool markOrigin;
  bool verbose;

  Options();
  void parse(const int argc, const char* argv[]);
  void report(ostream& s) const;
  static void usage();

    private:
  static const OptionSyntax syntax[];
};


/*********************************************************
* Geometry headers                                       *
*********************************************************/

const DOUBLE PI = 3.141592653589793;
const DOUBLE TWOPI = PI * 2.0;
const DOUBLE HALFPI = PI * 0.5;


class Point
{
    private:
  DOUBLE mX;
  DOUBLE mY;
  DOUBLE mR; // used to to store the length, first time it is asked
    public:
  Point(const DOUBLE _x = 0, const DOUBLE _y = 0);
  Point normalise(DOUBLE length = 1.0);
  Point operator+(const Point& p) const;
  Point operator-(const Point& p) const;
  DOUBLE operator*(const Point& p) const;
  Point operator*(const DOUBLE scalar) const;
  bool operator==(const Point& p) const;
  DOUBLE x() const; // x-coordinate
  DOUBLE y() const; // y-coordinate
  DOUBLE r(); // length/distance from origin
  DOUBLE fi() const; // direction from origin
  DOUBLE ySlope() const; // dy/dx
  DOUBLE xSlope() const; // dx/dy
  Point rotLeft() const;
  friend istream& operator>>(istream& is, Point& p);
};


istream& operator>>(istream& is, Point& p);


/*********************************************************
* Site headers                                           *
*********************************************************/

struct Site
{
  int size;
  Point* points;  // indexed from 0 to size-1
  Point* vectors; // indexed from 1 to size-1 
  DOUBLE weight;
  int id;
  Site();
  void erase();
  void setId(int _id);
  ~Site();
};


istream& operator>>(istream& is, Site& s);


/*********************************************************
* Highway map headers                                    *
*********************************************************/


struct ValueId {
  DOUBLE value;
  int id;
  ValueId(DOUBLE _value, int _id);
  bool operator<(const ValueId& vi) const;
};


class Canvas; 


class HighwayMap
{
    private:
  struct HighwaySection
  {
    Point p[2];
    Point invVector[2]; // 2x2 matrix row by row such that
                        // invVector * (p[1] - p[0]) = (1,0).
    DOUBLE speed;
    DOUBLE cotanOfEntry;
    std::vector<ValueId> ramps;
    HighwaySection(Point& p0, Point& p1, DOUBLE _speed);
    void calculateRamps(DOUBLE* t, Point& q);
    void calculateInterchange(
      HighwaySection& other, DOUBLE& tSelf, DOUBLE& tOther);
    Point rampLocation(const DOUBLE t) const;
  };   
  std::vector<HighwaySection> sections;

  struct HighwayNode
  {
    Point location;
    std::vector<ValueId> neighbours;
    DOUBLE* distances;
    HighwayNode(Point& _location, int nrSites);
    void connect(int nodeId, DOUBLE distance);
  };
  std::vector<HighwayNode> nodes;
  int nrPrimaryNodes;

  struct Cell
  {
    std::vector<int> nodes;
    std::vector<int> sections;
  };
  Cell*** cells; // usage: cells[siteId][gridRow][gridColumn]
  static const int gridWidth;

  struct Probe 
  {
    DOUBLE distance;
    int nodeId;
    int siteId;
    Probe(DOUBLE _distance, int _nodeId, int _siteId);
    bool operator<(const Probe& p) const;
  };
  DOUBLE distance(Point& p, HighwayNode& n, int siteId);
  DOUBLE distance(Point& p, HighwaySection& hw, int siteId);

    public:
  HighwayMap(std::vector<Site>& sites, std::vector<Site>& highways);
  DOUBLE distance(Point& p, int siteId);

  // for debugging purposes:
  void showNodes(Canvas& canvas);
};


/*********************************************************
* Distance headers                                       *
*********************************************************/

DOUBLE sqr(const DOUBLE x);


DOUBLE safeAcos(const DOUBLE x);


DOUBLE& keepMin(DOUBLE& x, const DOUBLE y);


DOUBLE& keepMax(DOUBLE& x, const DOUBLE y);


class Distance 
{
    public:
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void verifySite(Site& s) const;
};


class WeightedDistance: public Distance
{
    private:
  Distance* distance;
  DOUBLE weight;
    public:
  WeightedDistance(Distance* _distance, const DOUBLE _weight);
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void verifySite(Site& s) const;
};


class SumDistance: public Distance
{
    private:
  std::vector<WeightedDistance> distances;
    public:
  SumDistance* add(Distance* distance, const DOUBLE weight = 1.0);
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void verifySite(Site& s) const;
};


class EuclideanDistance: public Distance
{
    public:
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


class SquaredDistance: public Distance
{
    public:
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


class HighwayDistance: public Distance
{
    public:
  HighwayDistance();
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


class ManhattanDistance: public Distance
{
    public:
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


class MinDistance: public Distance
{
    public:
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


class ProductDistance: public Distance
{
    public:
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


class MaxDistance: public Distance
{
    public:
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


class LDistance: public Distance
{
    private:
  DOUBLE n;
  DOUBLE invn;
    public:
  LDistance(const DOUBLE _n);
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


class KarlsruheDistance: public Distance
{
    public:
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


class CityDistance: public Distance
{
    public:
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


class AzimuthDistance: public Distance
{
    public:
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


class LogRadiusDistance: public Distance
{
    public:
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


class OrbitOutDistance: public Distance
{
    public:
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


class OrbitInDistance: public Distance
{
    public:
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


class AngleDistance: public Distance
{
    public:
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void verifySite(Site& s) const;
};


class DetourDistance: public Distance
{
    public:
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void verifySite(Site& s) const;
};


class DilationDistance: public Distance
{
    public:
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void verifySite(Site& s) const;
};


class TurnDistance: public Distance
{
    public:
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void verifySite(Site& s) const;
};


class LeftTurnDistance: public Distance
{
    public:
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void verifySite(Site& s) const;
};


class DubinsDistance: public Distance
{
    public:
  virtual const string description() const;
  virtual bool supportsPolylines() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void verifySite(Site& s) const;
};


Distance* getDistanceObject(const string& name);


/*********************************************************
* Canvas headers                                         *
*********************************************************/

class Canvas
{
    private:
  const int nrRows;
  const int nrCols;
  RGB* pixels;
  RGB** rows;
  int penSize[nrPens];
  int* penMasks[nrPens];
    public:
  Canvas(const int _nrRows, const int _nrCols);
  const RGB& operator()(const int row, const int col) const;
  void shade(const int row, const int col, const RGB& colour);
  void draw(const int row, const int col, const int pen, const RGB& colour);
};


void drawSite(Canvas& c, const Site& s, const int pen, const RGB& colour);


/*********************************************************
* Owner headers                                          *
*********************************************************/

struct Owner // (Region ID)
{
  int primary;   // site number; -1 if none
  int secondary; // site number; -1 if none
  Owner(const int _primary = -1, const int _secondary = -1);
  void select(DOUBLE d[], const int nrSites);
  const bool operator==(const Owner& o) const;
  const bool operator!=(const Owner& o) const;
  const int representative(const int row, const int col) const;
};


/***********************************************************************

IMPLEMENTATION


**********************************************************
* Input/Output                                           *
*********************************************************/

const char* Complaint::errorLabel[4] = 
{ 
  "WARNING", 
  "Error",
  "Error", 
  "Internal error" 
};


Complaint::Complaint(ErrorLevel _errorLevel): 
  errorLevel(_errorLevel) 
{
};

  
Complaint::Complaint(
  ErrorLevel _errorLevel, 
  const string& expected, 
  const string& read
):
  errorLevel(_errorLevel)
{
  complaint << "expected \"" << expected << "\", but read \"" << read << "\".";
}


Complaint::~Complaint()
{
  cerr << endl << errorLabel[(int) errorLevel] << ": " << complaint.str() << endl;
  if (errorLevel == usageError) cerr << "Run vorosketch -? for usage." << endl;
  cerr << endl;
  if (errorLevel > warning) exit((int) errorLevel); 
}


template <typename T> Complaint& Complaint::operator<<(const T& t)
{
  complaint << t; return *this;
}


void makeLowerCase(string& s)
{
  for (int i = 0; i < s.length(); ++i) s[i] = tolower(s[i]);
}


DOUBLE strtofstop(const string& str)
{
  stringstream s(str);
  DOUBLE x;
  s >> x >> std::ws;
  if (!s.eof()) Complaint(userError)
    << "expected a floating-point number but read \"" << str << "\".";
  return x; 
}


int strtoistop(const string& str)
{
  stringstream s(str);
  int n;
  s >> n >> std::ws;
  if (!s.eof()) Complaint(userError)
    << "expected an integer but read \"" << str << "\".";
  return n; 
}


void reportProgress(DOUBLE x, string description)
{
  if (x >= 1)
    cerr << '\r' << description << ": done." << string(progressUnits - 3, ' ');
  else
  {
    cerr << '\r' << description << ": [";
    for (int i = 1; i < progressUnits; ++i)
    {
      if (x * progressUnits >= i)
        cerr << (i % 5 ? '-' : '+');
      else
        cerr << (i % 5 ? ' ' : ':');
    }
    cerr << "]";
  }
  cerr.flush();
}


RGB::RGB(const int _red, const int _green, const int _blue):
  mRed(_red),
  mGreen(_green),
  mBlue(_blue)
{};


RGB::RGB(const double _grey):
  mRed(_grey * 255),
  mGreen(_grey * 255),
  mBlue(_grey * 255)
{};


const int RGB::red() const
{
  return mRed;
}


const int RGB::green() const
{
  return mGreen;
}


const int RGB::blue() const
{
  return mBlue;
}


RGB& RGB::operator*=(const RGB& colour)
{
  mRed *= colour.mRed; mRed /= 255;
  mGreen *= colour.mGreen; mGreen /= 255;
  mBlue *= colour.mBlue; mBlue /= 255;
  return *this;
}


RGB RGB::black(0,0,0);
RGB RGB::grey(192,192,192);
RGB RGB::white(255,255,255);


RGB Trubetskoy[19] = {
  RGB(230,25,75), RGB(60,180,75), RGB(255,225,25), RGB(0,130,200),
  RGB(245,130,48), RGB(145,30,180), RGB(70,240,240), RGB(240,50,230),
  RGB(210,245,60), RGB(250,190,212), RGB(0,128,128), RGB(220,190,225),
  RGB(170,110,40), RGB(255,250,200), RGB(128,0,0), RGB(170,255,195),
  RGB(128,128,0), RGB(255,215,180), // RGB(0,0,128),  
  RGB(128,128,128)
};


BitmapFile::BitmapFile(ostream& _outputStream, const int _width, const int height):
  outputStream(_outputStream),
  width(_width),
  samplesYetToCome(_width * height),
  padding((4 - ((3*width) % 4)) % 4) 
{
  // write header:
  const int headerSize = 26;
  const int lineSize = width * 3 + padding;
  const int fileSize = headerSize + height * lineSize;
  writeInteger(0x42, 1);
  writeInteger(0x4D, 1);
  writeInteger(fileSize, 4);
  writeInteger(0, 4);
  writeInteger(headerSize, 4);
  writeInteger(12, 4);
  writeInteger(width, 2);
  writeInteger(height, 2);
  writeInteger(1, 2);
  writeInteger(24, 2);
}


BitmapFile& BitmapFile::writePixel(const RGB& colour) 
{
  if (--samplesYetToCome < 0)
    Complaint(internalError) << "attempt to write to finished bitmap file";
  writeInteger(colour.blue(), 1);
  writeInteger(colour.green(), 1);
  writeInteger(colour.red(), 1);
  if (samplesYetToCome % width == 0) writeInteger(0, padding);
  return *this;
}


BitmapFile::~BitmapFile() 
{ 
  if (samplesYetToCome > 0)
    Complaint(internalError) << "bitmap file incomplete";
}


/*********************************************************
* Options                                                *
*********************************************************/

// first declare some helpers:

const OptionSyntax* findOption(const OptionSyntax options[], string argv);

struct Argument 
// class to store options with arguments;
// different options must have unique ranks and ids
{
  int position; // position in the input   
  const OptionSyntax* key;
  string value;
  Argument(int _position, const OptionSyntax& _key, const string& _value);

  // functions to extract a number from the string value, and 
  // verify that the number is within given bounds:
  int intValue(const int min, const int max) const;
  DOUBLE floatValue(const DOUBLE min, const DOUBLE max) const;

  // comparison operator to sort by rank, and for options of equal rank,
  // by position in the input:
  bool operator<(const Argument& a) const;
};


// and here is the implementation:

OptionSyntax::OptionSyntax(
  int _rank, 
  string _id, 
  bool _requiresValue,
  bool _multipleAllowed
):
  rank(_rank), 
  id(_id), 
  requiresArgument(_requiresValue),
  multipleAllowed(_multipleAllowed)
{
};


// Options::Options(const int argc, const char* argv[]):
Options::Options():
  distance(new EuclideanDistance()),
  useHighways(false),
  weighting(Options::subtractive),
  equalityThreshold(0),
  segmentShortening(0),
  order(Options::first),
  drawSites(true),
  drawRegionBoundaries(false),
  nrColours(1),
  colours(&RGB::white),
  drawStandardContours(false),
  contourInterval(0),
  shadeInnerAreas(true),
  contourDistance(0),
  halfBrightDistance(-1),
  width(2000),
  markOrigin(false),
  verbose(false)
{};


void Options::parse(const int argc, const char* argv[])
{
  // collect command line arguments, check syntax
  std::vector< Argument > arguments;
  for (int i = 1; i < argc; ++i)
  {
    const OptionSyntax* o = findOption(syntax, string(argv[i]));
    if (!o) Complaint(usageError) << "unrecognised option \"" << argv[i] << "\".";
    if (o->requiresArgument && ++i == argc) 
      Complaint(usageError) << "missing value for \"" << o->id << "\" option.";
    arguments.push_back(Argument(arguments.size(), *o, argv[i]));           
  }

  // sort command line arguments, check for missing or double arguments
  std::sort(arguments.begin(), arguments.end());
  for (int i = 1; i < arguments.size(); ++i)
    if (
      arguments[i].key->rank == arguments[i-1].key->rank && 
      arguments[i].value != arguments[i-1].value &&
      !arguments[i].key->multipleAllowed
    )
      Complaint(usageError) << 
      "multiple values for \"" << arguments[i].key->id + "\" option (\"" << 
      arguments[i-1].value << "\" and \"" <<
      arguments[i].value << "\").";
 
  // process command line arguments
  for (std::vector< Argument >::const_iterator a = arguments.begin(); 
    a != arguments.end(); 
    ++a)
  {
    switch(a->key->id[1]) 
    {
      case 'm':
        delete distance;
        {
          string name(a->value);
          makeLowerCase(name);
          name = name + '+';
          int cut = name.find('+');
          if (cut == name.length()-1)
            distance = getDistanceObject(name.substr(0, cut));
          else
          {
            SumDistance* distances = new SumDistance();
            while (name.length() > 1)
            {
              distances->add(getDistanceObject(name.substr(0, cut)));
              name.erase(0, cut+1);
              cut = name.find('+');
            }
            distance = distances; 
          }
        }
        break;
      case 'd': weighting = divisive; contourDistance = 1; break;
      case '2': order = second; shadeInnerAreas = false; break;
      case 'f': order = farthest; shadeInnerAreas = false; break;
      case 'e': 
        if (order == second) 
          Complaint(warning) 
            << "-e option not supported for second-order Voronoi Diagrams.";
        else equalityThreshold = a->floatValue(0, 1) * 2.0 / width;
        break;
      case 's': 
        if (!distance->supportsPolylines())
          Complaint(warning) << "-s option irrelevant or not supported for " 
            << distance->description() << " distance.";
        else segmentShortening = a->floatValue(0, 1) * 2.0 / width;
        break;
      case 'a': drawSites = false; break;
      case 'b': drawRegionBoundaries = true; break;
      case 'c': nrColours = 19; colours = Trubetskoy; break;
      case 'u': drawStandardContours = true; break;
      case 'i': contourInterval = a->floatValue(0, 100); break;
      case 'g': 
        halfBrightDistance = a->floatValue(0, 100); 
        shadeInnerAreas = false; 
        break;
      case 'r': width = a->intValue(2, 5000); break;
      case 'v': verbose = true; break;
      case '?': usage(); break;
      case '+': markOrigin = true; break;
      default: Complaint(internalError) << "unrecognized option";
    } 
  }
  if (nrColours == 1) drawRegionBoundaries = true;
  if (verbose) report(cerr);
}


void Options::report(ostream& s) const
{
  cerr << "Order: ";
  if (order == first) cerr << "1 (regions determined by closest site)" << endl; 
  else if (order == second) cerr << "2 (regions determined by closest two sites)" << endl; 
  else if (order == farthest) cerr << "n-1 (regions determined by farthest site)" << endl; 
  else Complaint(internalError) << "order set incorrectly";

  cerr << "Distance measure: " 
    << distance->description() << endl;

  cerr << "Weighting: ";
  if (weighting == subtractive) cerr << "subtract weights" << endl; 
  else if (weighting == divisive)    cerr << "divide by weights" << endl; 
  else Complaint(internalError) << "weighting method set incorrectly";

  cerr << "Distances considered equal if difference at most: " 
    << equalityThreshold << endl;

  if (distance->supportsPolylines()) cerr << "Polyline sites shortened by: "
    << segmentShortening << endl;

  cerr << "Image size: " << width << 'x' << width << " pixels" << endl;

  cerr << "Origin: ";
  if (!markOrigin) cerr << "not ";
  cerr << "marked" << endl;

  cerr << "Sites: " ;
  if (drawSites == false) cerr << "not drawn" << endl;
  else if (nrColours == 1) cerr << "black" << endl; 
  else if (nrColours > 1) 
    cerr << "distinguished by up to " << nrColours << " colours" << endl; 
  else Complaint(internalError) << "colour scheme set incorrectly";

  cerr << "Regions: ";
  if (nrColours == 1) cerr << "white" << endl; 
  else if (drawSites) cerr << "in matching colours" << endl;
  else cerr << "coloured" << endl;

  cerr << "Unit distance contours: ";
  if (drawStandardContours)
  {
    cerr << "distance " << contourDistance << " marked by contour line";
    if (shadeInnerAreas) cerr << ", area within shaded";
    cerr << endl;
  }
  else if (shadeInnerAreas) 
    cerr << "area within distance " << contourDistance << " shaded" << endl;
  else cerr << "not shown" << endl;

  cerr << "Further contour lines: ";
  if (contourInterval > 0) 
    cerr << "at " << contourInterval << " intervals" << endl;
  else cerr << "none" << endl;

  cerr << "Shading: ";
  if (halfBrightDistance != -1) 
    cerr << "50% brightness at distance " << halfBrightDistance << endl;
  else if (shadeInnerAreas)
    cerr << "dark shade within distance " << contourDistance << endl;
  else 
    cerr << "none" << endl;
} 


void Options::usage() 
{
  cerr << vorosketchInfo << endl;
#ifdef INCLUDE_USER_DISTANCES
  cerr << "extended with user-defined distance measures" << endl;
#endif
  cerr << endl
       << "USAGE    vorosketch [<options>]" << endl
       << "INPUT    sites in file format described below" << endl
       << "OUTPUT   bmp image file" << endl 
       << endl
       << vorosketchOptions 
       << endl
       << vorosketchFileFormat
       << endl;
  exit(-1);
}


const OptionSyntax Options::syntax[17] = 
{
  OptionSyntax( 0, "-?"),
  OptionSyntax( 1, "-m", true),
  OptionSyntax( 2, "-d"),
  OptionSyntax( 3, "-2"),
  OptionSyntax( 4 ,"-f"),
  OptionSyntax( 5, "-r", true),
  OptionSyntax( 6, "-e", true),
  OptionSyntax( 7, "-s", true),
  OptionSyntax( 8, "-c"),
  OptionSyntax( 9, "-a"),
  OptionSyntax(10, "-b"),
  OptionSyntax(11, "-u"),
  OptionSyntax(12, "-i", true),
  OptionSyntax(13, "-g", true),
  OptionSyntax(14, "-+"),
  OptionSyntax(15, "-v"),
  OptionSyntax(99, "") // sentinel end
}; 


const OptionSyntax* findOption(const OptionSyntax options[], string argv)
{
  string lowerCaseArg(argv.substr(0,4)); makeLowerCase(lowerCaseArg);
  for (const OptionSyntax* o = &options[0]; ; ++o) 
  {
    if (o->id.empty()) return 0;
    if (lowerCaseArg == o->id.substr(0,4)) return o;
  }
}


Argument::Argument(int _position, const OptionSyntax& _key, const string& _value):
  position(_position), key(&_key), value(_value)
{};

 
int Argument::intValue(const int min, const int max) const
{
  int x = strtoistop(value);
  if (x < min || x > max) Complaint(usageError) 
    << "unexpected value \"" + value + "\" for \"" << key->id << "\" option.\n"
    << "Expected value between " << min << " and " << max << ".";
  return x;
}


DOUBLE Argument::floatValue(const DOUBLE min, const DOUBLE max) const
{
  DOUBLE x = strtofstop(value);
  if (x < min || x > max) Complaint(usageError) 
    << "unexpected value \"" + value + "\" for \"" << key->id << "\" option.\n"
    << "Expected value between " << min << " and " << max << ".";
  return x;
}


bool Argument::operator<(const Argument& a) const 
{ 
  if (key->rank < a.key->rank) return true;
  return (position < a.position);
}


Options options;


/*********************************************************
* Geometry                                               *
*********************************************************/

Point::Point(const DOUBLE _x, const DOUBLE _y):
  mX(_x), mY(_y), mR(-1)
{};


Point Point::normalise(DOUBLE length)
{
  if (length < 0) 
    Complaint(internalError) << "normalise used with negative length";
  DOUBLE factor = length / r();
  return Point(mX * factor, mY * factor);
}


Point Point::operator+(const Point& p) const
{
  return Point(mX + p.mX, mY + p.mY);
}


Point Point::operator-(const Point& p) const
{
  return Point(mX - p.mX, mY - p.mY);
}


DOUBLE Point::operator*(const Point& p) const
{
  return mX * p.mX + mY * p.mY;
}


Point Point::operator*(const DOUBLE scalar) const
{
  return Point(mX * scalar, mY * scalar);
}


bool Point::operator==(const Point& p) const
{
  return (mX == p.mX && mY == p.mY);
}


DOUBLE Point::x() const
{
  return mX;
}


DOUBLE Point::y() const
{
  return mY;
}


DOUBLE Point::r()
{
  if (mR < 0) mR = sqrt(mX * mX + mY * mY);
  return mR;  
}


DOUBLE Point::fi() const
{
  if (mX == 0 && mY == 0) return 0;
  return atan2(mY, mX);  
}


DOUBLE Point::ySlope() const
{
  return mY/mX;
}


DOUBLE Point::xSlope() const
{
  return mX/mY;
}


Point Point::rotLeft() const
{
  return Point(-mY,mX);
}


istream& operator>>(istream& is, Point& p)
{
  is >> p.mX >> p.mY;
  p.mR = -1;
  return is;
}


/*********************************************************
* Sites                                                  *
*********************************************************/

Site::Site():
  size(0),
  weight(0)
{};


void Site::erase()
{
  if (size > 0) delete [] points;
  if (size > 1) delete [] vectors;
  size = 0;
  weight = 0;
}


void Site::setId(int _id)
{
  id = _id;
}


Site::~Site()
{
  erase();
}


istream& operator>>(istream& is, Site& s)
{
  s.erase();
  is >> s.size;
  if (s.size < 1) Complaint(userError) << "Site without points detected in input";
  s.points = new Point[s.size];
  for (int i = 0; i < s.size; ++i) is >> s.points[i];
  if (s.size > 1 && options.segmentShortening != 0)
  {
    Point cutFromStart = 
      (s.points[0] - s.points[1]).normalise(options.segmentShortening);
    Point cutFromEnd   = 
      (s.points[s.size-1] - s.points[s.size-2]).normalise(options.segmentShortening);
    s.points[0]        = s.points[0]        - cutFromStart;
    s.points[s.size-1] = s.points[s.size-1] - cutFromEnd;
  }
  if (s.size > 1)
  {
    s.vectors = new Point[s.size];  
    for (int i = 1; i < s.size; ++i)
      s.vectors[i] = s.points[i] - s.points[i-1];
  }
  is >> s.weight;
  return is;
}


/*********************************************************
* Highway map                                            *
*********************************************************/


ValueId::ValueId(DOUBLE _value, int _id):
  value(_value),
  id(_id)
{};


bool ValueId::operator<(const ValueId& vi) const
{
  if (value < vi.value) return true;
  if (value > vi.value) return false;
  return (id < vi.id);
}


HighwayMap::HighwaySection::HighwaySection(Point& p0, Point& p1, DOUBLE _speed):
  speed(_speed),
  cotanOfEntry(1.0L / sqrt(sqr(_speed) - 1))
{
  p[0] = p0; 
  p[1] = p1;
  Point vector = p[1]-p[0];
  DOUBLE invSqLength = 1.0L / (vector * vector);
  invVector[0] = vector * invSqLength;
  invVector[1] = vector.rotLeft() * invSqLength;
}


void HighwayMap::HighwaySection::calculateRamps(DOUBLE* t, Point& q)
{
  Point pq = q - p[0]; 
  Point relativeLocation(invVector[0] * pq, invVector[1] * pq); 
  t[0] = relativeLocation.x() - abs(relativeLocation.y()) * cotanOfEntry;
  t[1] = relativeLocation.x() + abs(relativeLocation.y()) * cotanOfEntry;
}


void HighwayMap::HighwaySection::calculateInterchange(
  HighwayMap::HighwaySection& other, DOUBLE& tSelf, DOUBLE& tOther
)
{  
  tSelf = INF; tOther = INF;
  Point otherTranslated = other.p[0] - p[0];
  Point other0relative(invVector[0] * otherTranslated, invVector[1] * otherTranslated);
        otherTranslated = other.p[1] - p[0];
  Point other1relative(invVector[0] * otherTranslated, invVector[1] * otherTranslated);
  if (other0relative.y() == other1relative.y()) return;
  tOther = other0relative.y() / (other0relative.y() - other1relative.y());
  tSelf = other0relative.x() * (-tOther + 1.0) + other1relative.x() * tOther;
}


Point HighwayMap::HighwaySection::rampLocation(const DOUBLE t) const
{
  return p[0]*(-t+1.0) + p[1]*t;
}


HighwayMap::HighwayNode::HighwayNode(Point& _location, const int nrSites):
  location(_location),
  distances(new DOUBLE[nrSites])
{
  for (int i = 0; i < nrSites; ++i) distances[i] = INF;
}


void HighwayMap::HighwayNode::connect(int nodeId, DOUBLE distance)
{
  neighbours.push_back(ValueId(distance, nodeId));
}


const int HighwayMap::gridWidth = 100;


HighwayMap::Probe::Probe(DOUBLE _distance, int _nodeId, int _siteId):
  distance(_distance),
  nodeId(_nodeId),
  siteId(_siteId)
{};


bool HighwayMap::Probe::operator<(const HighwayMap::Probe& p) const
{
  if (distance > p.distance) return true;
  if (distance < p.distance) return false;
  return (siteId > p.siteId);
}


// distance via node
DOUBLE HighwayMap::distance(Point& p, HighwayNode& n, int siteId)
{
  return((p - n.location).r() + n.distances[siteId]);
}


// distance via highway section
DOUBLE HighwayMap::distance(Point& p, HighwaySection& hw, int siteId)
{
  DOUBLE d = INF;
  DOUBLE t[2];
  hw.calculateRamps(t, p);
  for (int i = 0; i < 2; ++i)
  {
    if (t[i] <= 0 || t[i] >= 1) continue;
    Point rampLocation = hw.rampLocation(t[i]);

    int j = 0;
    // find the first node in direction i; will simple linear search do?
    while (hw.ramps[j].value < t[i]) ++j; // safe bcs last ramp is at 1 >= t[i];

/* alternatively: use binary search:
    int maxj = hw.ramps.size()-1; // size() >= 2, so maxj >= 1
    while (j < maxj-1) // so maxj >= j+2
    {
      int midj = (j + maxj)/2;
      if (hw.ramps[midj].value < t[i]) j = midj; 
      else maxj = midj;
    } 
    // now ramps[j].value < t[i] <= ramps[j+1].value
    ++j;
*/
    // now ramps[j-1].value < t[i] <= ramps[j].value
    HighwayNode& n = nodes[hw.ramps[j-1+i].id];
    keepMin(d, 
      (p - rampLocation).r() + 
      (rampLocation - n.location).r() / hw.speed +
      n.distances[siteId]
    );
  }
  return d;
}


HighwayMap::HighwayMap(std::vector<Site>& sites, std::vector<Site>& highways)
{
  int nrSites = sites.size();
  int nrHighways = highways.size();

  std::priority_queue<Probe> queue;
  // create highway sections; create nodes for their endpoints, and
  // connect them to the other sections' endpoints
  for (int i = 0; i < nrHighways; ++i)
  {
    if (highways[i].size <= 1)
      Complaint(userError) << "Highway with only one point detected";
    if (highways[i].weight <= 1)
      Complaint(userError) << "Highways must have speed greater than 1";
    int nrNodesOnPreviousHighways = nodes.size();
    for (int j = 0; j < highways[i].size; ++j)
    {
      int h = nodes.size();
      nodes.push_back(HighwayNode(highways[i].points[j], nrSites));
      for (int g = (j == 0 ? h-1 : h-2); g >= 0; --g)
      {
        DOUBLE distance = (nodes[h].location - nodes[g].location).r();
        nodes[h].connect(g, distance);
        nodes[g].connect(h, distance);
      }
      if (j == 0) continue;
      sections.push_back(
        HighwaySection(
          highways[i].points[j-1], highways[i].points[j], highways[i].weight
        )
      );
      sections.back().ramps.push_back(ValueId(0, h-1));
      sections.back().ramps.push_back(ValueId(1, h  ));
    }
  }
  int nrHighwayEndpoints = nodes.size();

  // create vertices for the sites, connect them to the section
  // endpoints, and insert them in the queue for the SP computation
  for (int s = 0; s < sites.size(); ++s)
  {
    queue.push(Probe(0, nrHighwayEndpoints + s, s));
    nodes.push_back(HighwayNode(sites[s].points[0], nrSites));
    for (int g = 0; g < nrHighwayEndpoints; ++g)
      nodes.back().connect(g, (nodes.back().location - nodes[g].location).r());
  }
  nrPrimaryNodes = nodes.size();

  // create vertices for all highway intersections
  for (int h = 0; h < sections.size(); ++h)
  {
    HighwaySection& hs = sections[h];
    for (int g = 0; g < h; ++g)
    {
      HighwaySection& gs = sections[g];
      DOUBLE gt, ht; 
      gs.calculateInterchange(hs, gt, ht);
      if (ht <= 0 || ht >= 1 || gt <= 0 || gt >= 1) continue;
      int interchangeId = nodes.size();
      Point interchangeLocation = gs.rampLocation(gt);
      nodes.push_back(HighwayNode(interchangeLocation, nrSites));
      gs.ramps.push_back(ValueId(gt, interchangeId));
      hs.ramps.push_back(ValueId(ht, interchangeId));
    }
  }
  int nrInputNodes = nodes.size();

  // create vertices for all points where a ray from one of the 
  // sites or section endpoints reaches another highway section
  // under angle arccos(1/speed).
  for (int h = 0; h < sections.size(); ++h)
  {
    HighwaySection& hw = sections[h];
    for (int v = 0; v < nrPrimaryNodes; ++v)
    {
      if (h == v/2) continue; // v is one of the endpoints of hw
      DOUBLE t[2];
      hw.calculateRamps(t, nodes[v].location);
      for (int i = 0; i < 2; ++i)
      {
        if (t[i] <= 0 || t[i] >= 1) continue;
        Point rampLocation = hw.rampLocation(t[i]);
        DOUBLE distance = (rampLocation - nodes[v].location).r();
        int rampNodeId = nodes.size();
        nodes.push_back(HighwayNode(rampLocation, nrSites));
        nodes[v].connect(rampNodeId, distance);
        nodes[rampNodeId].connect(v, distance);
        hw.ramps.push_back(ValueId(t[i], rampNodeId));     
      }
    }
    sort(hw.ramps.begin(), hw.ramps.end()); // by t-value (first of Ramp)
    for (int i = 1; i < hw.ramps.size(); ++i)
    {
      DOUBLE distance = 
        (nodes[hw.ramps[i].id].location - nodes[hw.ramps[i-1].id].location).r()
        / hw.speed;
      nodes[hw.ramps[i-1].id].connect(hw.ramps[i].id, distance);
      nodes[hw.ramps[i].id].connect(hw.ramps[i-1].id, distance);
    }
  }

  // use Dijkstra's algorithm to compute first and second-closest site
  // to each node
  while (!queue.empty())
  {
    Probe p = queue.top(); queue.pop();
    HighwayNode& n = nodes[p.nodeId];
    if (p.distance >= n.distances[p.siteId]) continue;
    n.distances[p.siteId] = p.distance;
    for (int i = 0; i < n.neighbours.size(); ++i)
      queue.push(
        Probe(
          p.distance + n.neighbours[i].value, n.neighbours[i].id, p.siteId 
        )
      );
  }

  Cell* allcells = new Cell[nrSites * gridWidth * gridWidth];
  Cell** allrows = new Cell*[nrSites * gridWidth];
  for (int i = 0; i < nrSites * gridWidth; ++i) 
  {
    allrows[i] = allcells;
    allcells += gridWidth;
  }
  cells = new Cell**[nrSites];
  for (int i = 0; i < nrSites; ++i)
  {
    cells[i] = allrows;
    allrows += gridWidth;
  }

  int totalRegisteredNodes = 0;
  int totalRegisteredSections = 0;
  const DOUBLE cellDiameter = 2.0/gridWidth * sqrt(2.0);
  DOUBLE* nodeDistance[nrSites];
  for (int i = 0; i < nrSites; ++i) 
    nodeDistance[i] = new DOUBLE[nrPrimaryNodes];
  DOUBLE* sectionDistance[nrSites];
  for (int i = 0; i < nrSites; ++i) 
    sectionDistance[i] = new DOUBLE[sections.size()];
  DOUBLE upperBound[nrSites];

  for (int gridRow = 0; gridRow < gridWidth; ++gridRow)
    for (int gridCol = 0; gridCol < gridWidth; ++gridCol)
    {
      Point cellCentre(
        (2.0*gridCol+1.0)/gridWidth-1.0, 
        (2.0*gridRow+1.0)/gridWidth-1.0
      );
      for (int siteId = 0; siteId < nrSites; ++siteId)
      {
        upperBound[siteId] = INF;
        for (int i = 0; i < nrPrimaryNodes; ++i)
        {
          nodeDistance[siteId][i] = distance(cellCentre, nodes[i], siteId);
          keepMin(upperBound[siteId], nodeDistance[siteId][i]);
        }
        for (int h = 0; h < sections.size(); ++h)
        {
          sectionDistance[siteId][h] = distance(cellCentre, sections[h], siteId);
          keepMin(upperBound[siteId], sectionDistance[siteId][h]);
        }
      }
      DOUBLE firstUpperBound = INF;
      DOUBLE secondUpperBound = INF;
      for (int siteId = 0; siteId < nrSites; ++siteId)
      {
        keepMin(secondUpperBound, upperBound[siteId]);
        if (secondUpperBound < firstUpperBound) 
            swap(firstUpperBound, secondUpperBound);
      }
      DOUBLE mapUpperBound;
      switch (options.order) 
      {
        case Options::first:    
          mapUpperBound = firstUpperBound + options.equalityThreshold;
          break;
        case Options::second:
          mapUpperBound = secondUpperBound + options.equalityThreshold;
          break;
        case Options::farthest:
          mapUpperBound = INF;
          break;
        default: 
          Complaint(internalError) << 
            "unknown diagram order in HighwayMap constructor";
      }

      for (int siteId = 0; siteId < nrSites; ++siteId)
      {    
        Cell& cell = cells[siteId][gridRow][gridCol];
        DOUBLE cellUpperBound = 
          cellDiameter + min(mapUpperBound, upperBound[siteId]);

        // distance(cellCentre, s) + cellDiameter/2 is an upper bound
        // for routes via n, 
        //   distance from centre via n - cellDiameter/2 is a lower bound,
        //   so need to be considered only if 
        //   distance from centre via n - cellDiameter/2 < 
        //     distance(cellCentre, s) + cellDiameter/2         <=>
        //   distance from centre via n < distance(cellCentre, s) + cellDiameter

        for (int i = 0; i < nrPrimaryNodes; ++i)
        {
          if (nodeDistance[siteId][i] > cellUpperBound) continue;
          cell.nodes.push_back(i);
          ++totalRegisteredNodes;
        }
        for (int h = 0; h < sections.size(); ++h)
        {
          if (sectionDistance[siteId][h] > cellUpperBound) continue;
          cell.sections.push_back(h);
          ++totalRegisteredSections;
        }
      }
    }

  for (int i = 0; i < nrSites; ++i) delete[] nodeDistance[i];
  for (int i = 0; i < nrSites; ++i) delete[] sectionDistance[i];

/*
  if (options.verbose)
  {
    cerr << "\r"
         << "Highway map grid size: " 
         << gridWidth << " x " << gridWidth << " cells" << endl
         << "Average number of registered nodes per site per cell: "
         << (DOUBLE) totalRegisteredNodes / (nrSites * gridWidth * gridWidth) << endl
         << "Average number of registered highway sections per site per cell: "
         << (DOUBLE) totalRegisteredSections / (nrSites * gridWidth * gridWidth) << endl
         << "Preprocessing input..."; 
    cerr.flush();
  }
*/
}


DOUBLE HighwayMap::distance(Point& p, const int siteId)
{
  int gridRow = (p.y() + 1.0) * 0.5 * gridWidth; 
  if (gridRow < 0) gridRow = 0; else 
  if (gridRow >= gridWidth) gridRow = gridWidth-1;
  int gridCol = (p.x() + 1.0) * 0.5 * gridWidth;
  if (gridCol < 0) gridCol = 0; else 
  if (gridCol >= gridWidth) gridCol = gridWidth-1;

  Cell& cell = cells[siteId][gridRow][gridCol];
  DOUBLE d = INF;

  for (
    std::vector<int>::iterator i = cell.nodes.begin(); 
    i != cell.nodes.end(); 
    ++i
  )
    keepMin(d, distance(p, nodes[*i], siteId));

  for (
    std::vector<int>::iterator i = cell.sections.begin(); 
    i != cell.sections.end(); 
    ++i
  )
    keepMin(d, distance(p, sections[*i], siteId));

  return d;
}


void HighwayMap::showNodes(Canvas& canvas)
{
  for (int i = 0; i < nodes.size(); ++i)
  {
    Point pixel(
      0.5 * (nodes[i].location.x() + 1.0) * options.width + 0.5,
      0.5 * (nodes[i].location.y() + 1.0) * options.width + 0.5
    );
    canvas.draw(pixel.y(), pixel.x(), bisectorPen, RGB::black);
  }
}


HighwayMap* highwayMap = NULL;


/*********************************************************
* Distance                                               *
*********************************************************/

DOUBLE sqr(const DOUBLE x)
{
  return x * x;
}


DOUBLE safeAcos(const DOUBLE x)
{
  if (x >= 1) return 0;
  if (x <= -1) return PI;
  return acos(x);
}


DOUBLE& keepMin(DOUBLE& x, const DOUBLE y)
{
  if (y < x) x = y;
  return x;
}


DOUBLE& keepMax(DOUBLE& x, const DOUBLE y)
{
  if (y > x) x = y;
  return x;
}


const string Distance::description() const
{
  Complaint(internalError) 
    << "virtual function Distance::description() called";
  return "abstract";
}


bool Distance::supportsPolylines() const
{
  Complaint(internalError) 
    << "virtual function Distance::supportsPolylines() called";
  return false;
}


DOUBLE Distance::operator()(Point& p, const Site& s) const
{
  Complaint(internalError) 
    << "virtual function Distance::operator()() called";
  return 0;
}


void Distance::verifySite(Site& s) const
{
  if (!supportsPolylines() && s.size > 1) Complaint(userError) 
    << "input contains polyline site, "
    << "for which the " << description() << " distance has not been implemented";

  static bool zeroWeightWarningGiven = false;
  if (options.weighting == Options::divisive && s.weight == 0 && !zeroWeightWarningGiven) 
  {
    Complaint(warning) << "sites with zero weight have empty regions";
    zeroWeightWarningGiven = true;
  }
}


WeightedDistance::WeightedDistance(Distance* _distance, const DOUBLE _weight):
  distance(_distance),
  weight(_weight)
{};


const string WeightedDistance::description() const
{
  stringstream s;
  if (weight != 1) s << weight << '*';
  s << distance->description();
  return s.str();
}


bool WeightedDistance::supportsPolylines() const
{
  return distance->supportsPolylines();
}


DOUBLE WeightedDistance::operator()(Point& p, const Site& s) const
{
  return weight * (*distance)(p, s);
}


void WeightedDistance::verifySite(Site& s) const
{
  distance->verifySite(s);
}


SumDistance* SumDistance::add(Distance* distance, const DOUBLE weight)
{
  distances.push_back(WeightedDistance(distance, weight));
  return this;
}


const string SumDistance::description() const
{
  stringstream s;
  s << distances[0].description();
  for (int i = 1; i < distances.size(); ++i)
    s << " + " << distances[i].description();
  return s.str();
}


bool SumDistance::supportsPolylines() const
{
  bool s = true;
  for (int i = 0; i < distances.size(); ++i)
    s = s && distances[i].supportsPolylines();
  return s;
}


DOUBLE SumDistance::operator()(Point& p, const Site& s) const
{
  DOUBLE distance = 0;
  for (int i = 0; i < distances.size(); ++i)
  {
    distance += distances[i](p, s);
    if (distance == INF) return INF;
  }
  return distance;
}


void SumDistance::verifySite(Site& s) const
{
  for (int i = 0; i < distances.size(); ++i)
    distances[i].verifySite(s);
}


const string EuclideanDistance::description() const
{
  return "Euclidean";
}


bool EuclideanDistance::supportsPolylines() const
{
  return true;
}


DOUBLE EuclideanDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];
  DOUBLE sqDistance = v * v;
  for (int i = 1; i < s.size; ++i) 
  {
    // calculate x-coordinate if axis system is rotated
    // and translated such that the segment points to the right,
    // and scaled with factor s.vectors[i].r():
    DOUBLE x = s.vectors[i] * v;
    // check if point lies vertically above or below the segment
    if (x > 0 && x < sqr(s.vectors[i].r()))
    {
      // calculate y-coordinate under same transformation
      DOUBLE y = s.vectors[i].rotLeft() * v;
      keepMin(sqDistance, sqr(y / s.vectors[i].r()));
    }
    v = p - s.points[i];
    keepMin(sqDistance, v * v);
  }
  return sqrt(sqDistance);
}


const string SquaredDistance::description() const
{
  return "squared Euclidean";
}


bool SquaredDistance::supportsPolylines() const
{
  return true;
}


DOUBLE SquaredDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];
  DOUBLE sqDistance = v * v;
  for (int i = 1; i < s.size; ++i) 
  {
    // calculate x-coordinate if axis system is rotated
    // and translated such that the segment points to the right,
    // and scaled with factor s.vectors[i].r():
    DOUBLE x = s.vectors[i] * v;
    // check if point lies vertically above or below the segment
    if (x > 0 && x < sqr(s.vectors[i].r()))
    {
      // calculate y-coordinate under same transformation
      DOUBLE y = s.vectors[i].rotLeft() * v;
      keepMin(sqDistance, sqr(y / s.vectors[i].r()));
    }
    v = p - s.points[i];
    keepMin(sqDistance, v * v);
  }
  return sqDistance;
}


HighwayDistance::HighwayDistance()
{
  options.useHighways = true;
}


const string HighwayDistance::description() const
{
  return "Highway";
}


bool HighwayDistance::supportsPolylines() const
{
  return false;
}


DOUBLE HighwayDistance::operator()(Point& p, const Site& s) const
{
  return highwayMap->distance(p, s.id);
}


const string ManhattanDistance::description() const
{
  return "Manhattan (L1)";
}


bool ManhattanDistance::supportsPolylines() const
{
  return true;
}


DOUBLE ManhattanDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];
  DOUBLE distance = abs(v.x()) + abs(v.y());
  for (int i = 1; i < s.size; ++i)  
  {
    if (abs(s.vectors[i].x()) > abs(s.vectors[i].y()))
    {
      // test if p.x lies between s.points[i-1].x and s.points[i].x,
      // works regardless of which endpoint has the smaller x-coordinate
      if ((v.x() > 0) == (v.x() < s.vectors[i].x()))
        keepMin(distance, abs(v.y() - v.x() * s.vectors[i].ySlope()));
    }
    else
    {
      if ((v.y() > 0) == (v.y() < s.vectors[i].y()))
        keepMin(distance, abs(v.x() - v.y() * s.vectors[i].xSlope()));
    }
    v = p - s.points[i];
    keepMin(distance, abs(v.x()) + abs(v.y()));
  }
  return distance;
}


const string MinDistance::description() const
{
  return "Minimum (L-inf)";
}


bool MinDistance::supportsPolylines() const
{
  return false;
}


DOUBLE MinDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];
  return min(abs(v.x()), abs(v.y()));
}


const string ProductDistance::description() const
{
  return "Product (bounding-box area)";
}


bool ProductDistance::supportsPolylines() const
{
  return false;
}


DOUBLE ProductDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];
  return abs(v.x()*v.y());
}


const string MaxDistance::description() const
{
  return "Maximum (Linf)";
}


bool MaxDistance::supportsPolylines() const
{
  return false;
}


DOUBLE MaxDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];
  return max(abs(v.x()), abs(v.y()));
}


LDistance::LDistance(const DOUBLE _n):
  n(_n)
{
  if (n == 0) Complaint(userError) << "L0-distance is undefined";
  invn = ((DOUBLE) 1.0) / n;
}


const string LDistance::description() const
{
  stringstream s;
  s << "L" << n;
  return s.str();
}


bool LDistance::supportsPolylines() const
{
  return false;
}


DOUBLE LDistance::operator()(Point& p, const Site& s) const
{
  if (n < 0 && (p.x() == s.points[0].x() || p.y() == s.points[0].y())) 
    return 0;
  return pow(
    pow(abs(p.x() - s.points[0].x()), n) +
    pow(abs(p.y() - s.points[0].y()), n),
  invn);
}


const string KarlsruheDistance::description() const
{
  return "Karlsruhe";
}


bool KarlsruheDistance::supportsPolylines() const
{
  return false;
}


DOUBLE KarlsruheDistance::operator()(Point& p, const Site& s) const
{
  DOUBLE angleDistance = safeAcos((p * s.points[0]) / (p.r() * s.points[0].r()));
  if (angleDistance > 2) return p.r() + s.points[0].r();
  return 
    angleDistance * min(p.r(), s.points[0].r()) + 
    abs(p.r() - s.points[0].r());
}


const string CityDistance::description() const
{
  return "city";
}


bool CityDistance::supportsPolylines() const
{
  return false;
}


DOUBLE CityDistance::operator()(Point& p, const Site& s) const
{
  if (p.r() == 0 || s.points[0].r() == 0) return INF;
  DOUBLE angleDistance = safeAcos((p * s.points[0]) / (p.r() * s.points[0].r()));
  DOUBLE radialDistance = log(p.r() / s.points[0].r());
  return sqrt(sqr(angleDistance) + sqr(radialDistance));
}


const string AzimuthDistance::description() const
{
  return "Azimuth";
}


bool AzimuthDistance::supportsPolylines() const
{
  return false;
}


DOUBLE AzimuthDistance::operator()(Point& p, const Site& s) const
{
  if (p.r() == 0 || s.points[0].r() == 0) return 0;
  return safeAcos((p * s.points[0]) / (p.r() * s.points[0].r()));
}


const string LogRadiusDistance::description() const
{
  return "LogRadius";
}


bool LogRadiusDistance::supportsPolylines() const
{
  return false;
}


DOUBLE LogRadiusDistance::operator()(Point& p, const Site& s) const
{
  if (p.r() == 0 || s.points[0].r() == 0) return INF;
  return abs(log(p.r() / s.points[0].r()));
}


const string OrbitOutDistance::description() const
{
  return "Orbit Out";
}


bool OrbitOutDistance::supportsPolylines() const
{
  return false;
}


DOUBLE OrbitOutDistance::operator()(Point& p, const Site& s) const
{
  return (
    ((DOUBLE) 1.0) / s.points[0].r() - 
    ((DOUBLE) 2.0) / (s.points[0].r() + (p - s.points[0]).r() + p.r())
  );
}


const string OrbitInDistance::description() const
{
  return "Orbit In";
}


bool OrbitInDistance::supportsPolylines() const
{
  return false;
}


DOUBLE OrbitInDistance::operator()(Point& p, const Site& s) const
{
  return (
    ((DOUBLE) 1.0) / p.r() - 
    ((DOUBLE) 2.0) / (s.points[0].r() + (p - s.points[0]).r() + p.r())
  );
}


const string AngleDistance::description() const
{
  return "Angle";
}


bool AngleDistance::supportsPolylines() const
{
  return true;
}


DOUBLE AngleDistance::operator()(Point& p, const Site& s) const
{
  if (s.points[0] == p) return (DOUBLE) 1.0 / TWOPI;
  DOUBLE angle = (s.points[0] - p).fi();
  DOUBLE minAngle = angle;
  DOUBLE maxAngle = angle;
  for (int i = 1; i < s.size; ++i)
  {
    if (s.points[i] == p) return (DOUBLE) 1.0 / TWOPI;
    DOUBLE prevAngle = angle;
    angle = (s.points[i] - p).fi();
    DOUBLE dAngle = fmod(angle - prevAngle + TWOPI * i, TWOPI);
    if (dAngle < PI) 
    {
      angle = prevAngle + dAngle;
      keepMax(maxAngle, angle);
    }      
    else // dAngle >= PI
    {
      angle = prevAngle + (dAngle - TWOPI);
      keepMin(minAngle, angle);
    }
  }
  if (maxAngle == minAngle) return INF;
  if (maxAngle - minAngle > TWOPI) return (DOUBLE) 1.0 / TWOPI;
  return (DOUBLE) 1.0 / (maxAngle - minAngle);
}


void AngleDistance::verifySite(Site& s) const
{
  static bool pointSiteWarningGiven = false;
  if (s.size == 1 && !pointSiteWarningGiven)
  {
    Complaint(warning) 
      << "input contains single-point site, "
      << "for which the angle distance is meaningless";
    pointSiteWarningGiven = true;
  }
  Distance::verifySite(s);
}


const string DetourDistance::description() const
{
  return "Detour";
}


bool DetourDistance::supportsPolylines() const
{
  return true;
}


DOUBLE DetourDistance::operator()(Point& p, const Site& s) const
{
  if (s.size == 1) return (s.points[0] - p).r() * 2.0;
  DOUBLE distance = INF;
  for (int i = 1; i < s.size; ++i)
    keepMin(distance, 
      (s.points[i-1] - p).r() + (p - s.points[i]).r() - s.vectors[i].r()
    );
  return distance;
}


void DetourDistance::verifySite(Site& s) const
{
  static bool pointSiteWarningGiven = false;
  if (s.size == 1 && !pointSiteWarningGiven)
  {
    Complaint(warning) 
      << "single-point site interpreted as line segment of zero length";
    pointSiteWarningGiven = true;
  }
  Distance::verifySite(s);
}


const string DilationDistance::description() const
{
  return "Dilation";
}


bool DilationDistance::supportsPolylines() const
{
  return true;
}


DOUBLE DilationDistance::operator()(Point& p, const Site& s) const
{
  DOUBLE distance = INF;
  for (int i = 1; i < s.size; ++i)
    keepMin(distance, 
      ((s.points[i-1] - p).r() + (p - s.points[i]).r()) / s.vectors[i].r() - 1.0
    );
  return distance;
}


void DilationDistance::verifySite(Site& s) const
{
  static bool pointSiteWarningGiven = false;
  if (s.size == 1 && !pointSiteWarningGiven)
  {
    Complaint(warning) 
      << "input contains single-point site, "
      << "for which the dilation distance is meaningless";
    pointSiteWarningGiven = true;
  }
  Distance::verifySite(s);
}


const string TurnDistance::description() const
{
  return "Turn (left and right)";
}


bool TurnDistance::supportsPolylines() const
{
  return false;
}


DOUBLE TurnDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];
  return safeAcos(s.vectors[1] * v / (v.r() * s.vectors[1].r()));
}


void TurnDistance::verifySite(Site& s) const
{
  if (s.size == 1) Complaint(userError) 
    << "input contains single-point site, "
    << "for which the turn distance is undefined";
  if (s.size > 2) Complaint(userError) 
    << "input contains polyline site, "
    << "for which the turn distance is undefined";
  static bool zeroWeightWarningGiven = false;
  if (options.weighting == Options::divisive && s.weight == 0 && !zeroWeightWarningGiven) 
  {
    Complaint(warning) << "sites with zero weight have empty regions";
    zeroWeightWarningGiven = true;
  }
}


const string LeftTurnDistance::description() const
{
  return "Left Turn";
}


bool LeftTurnDistance::supportsPolylines() const
{
  return false;
}


DOUBLE LeftTurnDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];
  if (v.x() == 0 && v.y() == 0) return 0;
  // rotate such that s.vectors[1] points in direction 0
  // (it also scales, but that does not matter)
  Point n(s.vectors[1] * v, s.vectors[1].rotLeft() * v);
  DOUBLE angle = atan2(n.y(), n.x());
  return (angle >= 0 ? angle : angle + TWOPI);
}


void LeftTurnDistance::verifySite(Site& s) const
{
  if (s.size == 1) Complaint(userError) 
    << "input contains single-point site, "
    << "for which the left-turn distance is undefined";
  if (s.size > 2) Complaint(userError) 
    << "input contains polyline site, "
    << "for which the left-turn distance is undefined";
  static bool zeroWeightWarningGiven = false;
  if (options.weighting == Options::divisive && s.weight == 0 && !zeroWeightWarningGiven) 
  {
    Complaint(warning) << "sites with zero weight have empty regions";
    zeroWeightWarningGiven = true;
  }
}


const string DubinsDistance::description() const
{
  return "Dubins";
}


bool DubinsDistance::supportsPolylines() const
{
  return false;
}


DOUBLE DubinsDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];

  // if p is at the starting point, return 0:
  if (v.x() == 0 && v.y() == 0) return 0;

  // get inverse turning radius:
  DOUBLE invr = ((DOUBLE) 1.0) / s.vectors[1].r();

  // rotate such that s.vectors[1] points in direction 0,
  // then reflect such that destination is not below x-axis
  Point n(
    s.vectors[1] * v * invr, 
    abs(s.vectors[1].rotLeft() * v * invr)
  );

  // the same in radius units:
  Point nu = n * invr;

  // get location relative to turning circle centre
  Point ru = nu - Point(0, 1);
 
  if (ru * ru < 1)
  { // within turning circle
    // get location relative to opposite turning circle centre
    ru = nu - Point(0, -1);
    // get distance to opposite turning circle centre
    DOUBLE ruSqLength = ru * ru;
    DOUBLE ruLength = sqrt(ruSqLength);
    return s.vectors[1].r() * (
      // right turn if destination would be on y-axis:
      safeAcos((DOUBLE) 0.25*ruLength + (DOUBLE) 0.75/ruLength) +
      // further right turn:
      atan2(ru.x(), ru.y()) +
      // left turn:
      TWOPI - safeAcos((DOUBLE) 1.25 - (DOUBLE) 0.25*ruSqLength)
    );
  }
  else
  { // outside turning circle
    DOUBLE ruSqLength = ru * ru;
    DOUBLE ruLength = sqrt(ruSqLength);
    return s.vectors[1].r() * (
      // direction relative to turning circle centre, to be modified such
      atan2(-ru.x(), ru.y()) + HALFPI +
      // additional angle because of approach from tangent point
      HALFPI - safeAcos((DOUBLE) 1.0 / ruLength) +
      // straight part:
      sqrt(ruSqLength - 1)
    );
  }
}


void DubinsDistance::verifySite(Site& s) const
{
  if (s.size == 1) Complaint(userError) 
    << "input contains single-point site, "
    << "for which the Dubins distance is undefined";
  if (s.size > 2) Complaint(userError) 
    << "input contains polyline site, "
    << "for which the Dubins distance is undefined";
  static bool zeroWeightWarningGiven = false;
  if (options.weighting == Options::divisive && s.weight == 0 && !zeroWeightWarningGiven) 
  {
    Complaint(warning) << "sites with zero weight have empty regions";
    zeroWeightWarningGiven = true;
  }
}


#ifdef INCLUDE_USER_DISTANCES
#include "userdistances.cpp"
#endif


Distance* getDistanceObject(const string& name)
{
  DOUBLE weight = 1;
  string dname = name;
  int star = dname.find('*');
  if (star < dname.length())
  {
    weight = strtofstop(dname.substr(0, star));
    dname = dname.substr(star+1);
    if (dname.empty()) 
      Complaint(usageError) << "missing distance name in \"" << name << "\"";
  }
  Distance* distance = NULL;
  if (dname == "euclidean" || dname == "l2") distance = new EuclideanDistance(); else
  if (dname == "squared") distance = new SquaredDistance(); else
  if (dname == "highway") distance = new HighwayDistance(); else
  if (dname == "manhattan" || dname == "l1") distance = new ManhattanDistance(); else
  if (dname == "karlsruhe") distance = new KarlsruheDistance(); else
  if (dname == "city") distance = new CityDistance(); else
  if (dname == "azimuth") distance = new AzimuthDistance(); else
  if (dname == "logradius") distance = new LogRadiusDistance(); else
  if (dname == "koeln") distance = (new SumDistance())
    ->add(new AzimuthDistance())->add(new LogRadiusDistance());
  else
  if (dname == "orbitout") distance = new OrbitOutDistance(); else
  if (dname == "orbitin") distance = new OrbitInDistance(); else
  if (dname == "angle") distance = new AngleDistance(); else
  if (dname == "detour") distance = new DetourDistance(); else
  if (dname == "dilation") distance = new DilationDistance(); else
  if (dname == "turn") distance = new TurnDistance(); else
  if (dname == "leftturn") distance = new LeftTurnDistance(); else
  if (dname == "dubins") distance = new DubinsDistance(); else
  if (dname == "min") distance = new MinDistance(); else
  if (dname == "bbox") distance = new ProductDistance(); else
  if (dname == "max") distance = new MaxDistance(); else
#ifdef INCLUDE_USER_DISTANCES
  MAKEUSERDISTANCE else
#endif
  if (dname[0] == 'l') {
    DOUBLE p = strtofstop(dname.substr(1));
    if (p == 0) distance = new ProductDistance();
    else distance = new LDistance(p);
  }
  else 
    Complaint(usageError) << "unrecognised distance name \"" << dname << "\"";
  if (weight == 1)
    return distance;
  else
    return new WeightedDistance(distance, weight);
}


/*********************************************************
* Canvas                                                 *
*********************************************************/


Canvas::Canvas(const int _nrRows, const int _nrCols):
  nrRows(_nrRows + strokeWidth[nrPens-1]),
  nrCols(_nrCols + strokeWidth[nrPens-1]),  
  pixels(new RGB[nrRows * nrCols]),
  rows(new RGB*[nrRows])
{
  // let rows[i] point to first pixel in output colum in row i
  rows[0] = pixels + strokeWidth[nrPens-1]/2;
  for (int i = 1; i < nrRows; ++i) rows[i] = rows[i-1] + nrCols;

  // reindex rows so that 0 is the first output row
  rows += strokeWidth[nrPens-1]/2;
   
  // calculate upper bound on total pen size
  int totalPenSize = 0;
  for (int i = 0; i < nrPens; ++i) totalPenSize += strokeWidth[i]*strokeWidth[i];

  // reserve space for pen masks
  int* penMask = new int[totalPenSize];

  // calculate pen masks
  for (int i = 0; i < nrPens; ++i)
  {
    penSize[i] = 0;
    penMasks[i] = penMask;
    int radius = strokeWidth[i]/2;
    DOUBLE sqRadius = sqr(radius);
    for (int row = -radius; row < radius; ++row)
      for (int col = -radius; col < radius; ++col)
        if (sqr(0.5 + (DOUBLE) row) + sqr(0.5 + (DOUBLE) col) <= sqRadius)
        {
          *(penMask++) = row * nrCols + col;
          ++penSize[i];
        }
  }
}
  

const RGB& Canvas::operator()(const int row, const int col) const
{
  return rows[row][col];
}


void Canvas::shade(const int row, const int col, const RGB& colour)
{
  rows[row][col] *= colour;
}


void Canvas::draw(const int row, const int col, const int pen, const RGB& colour)
{
  if (row < 0 || row >= options.width || col < 0 || col > options.width) return;
  RGB* centre = rows[row] + col;
  int* penMask = penMasks[pen];
  for (int i = penSize[pen]-1; i >= 0; --i) centre[penMask[i]] = colour;
}


void drawSite(Canvas& c, const Site& s, const int pen, const RGB& colour)
{
  // calculate pixel coordinates
  Point* corners = new Point[s.size];
  for (int i = 0; i < s.size; ++i)
    corners[i] = Point(
      0.5 * (s.points[i].x() + 1.0) * options.width + 0.5,
      0.5 * (s.points[i].y() + 1.0) * options.width + 0.5
    );
  if (s.size == 1) 
  {
    c.draw((int) corners[0].y(), (int) corners[0].x(), pen, colour);
  }
  else for (int i = 1; i < s.size; ++i)
    {
      Point vector = corners[i] - corners[i-1];
      if (abs(vector.x()) > abs(vector.y()))
      {
        int col = corners[i-1].x();
        int dCol = (vector.x() < 0 ? -1 : 1);
        int tCol = corners[i].x();
        DOUBLE dRow = vector.ySlope() * dCol;
        DOUBLE row = corners[i-1].y() - (corners[i-1].x() - col) * vector.ySlope();
        for (int j = abs(tCol - col); j >= 0; --j, col += dCol, row += dRow)
          c.draw((int) row, col, pen, colour);
      }
      else
      {
        int row = corners[i-1].y();
        int dRow = (vector.y() < 0 ? -1 : 1);
        int tRow = corners[i].y();
        DOUBLE dCol = vector.xSlope() * dRow;
        DOUBLE col = corners[i-1].x() - (corners[i-1].y() - row) * vector.xSlope();
        for (int j = abs(tRow - row); j >= 0; --j, col += dCol, row += dRow)
          c.draw(row, (int) col, pen, colour);
      }
    }
  delete [] corners;
}


/*********************************************************
* Owner                                                  *
*********************************************************/

Owner::Owner(const int _primary, const int _secondary):
  primary(_primary),
  secondary(_secondary)
{};


void Owner::select(DOUBLE d[], const int nrSites)
{
  DOUBLE sign = ( options.order == Options::farthest ? -1 : 1 );
  primary = -1;
  secondary = -1;
  if (nrSites < 1) return;
  primary = 0;
  if (nrSites < 2) return;

  // find the two best sites:
  secondary = 1;
  if (sign * d[primary] > sign * d[secondary]) swap(primary, secondary);
  for (int i = 2; i < nrSites; ++i)
  {
    if (sign * d[secondary] > sign * d[i]) secondary = i;
    if (sign * d[primary] > sign * d[secondary]) swap(primary, secondary);
  }
  
  if (options.order != Options::second)
  {
    // throw the 2nd site out if not within equality threshold of the 1st
    if (sign * d[secondary] > sign * d[primary] + options.equalityThreshold)
      secondary = -1;
  }

  // always list the site with the smallest id first:
  if (secondary != -1 && primary > secondary) swap(primary, secondary);
}


const bool Owner::operator==(const Owner& o) const
{
  return primary == o.primary && secondary == o.secondary;
}


const bool Owner::operator!=(const Owner& o) const
{
  return !operator==(o);
}


const int Owner::representative(const int row, const int col) const
{
  if (secondary == -1) return primary;
#ifdef CHECKERBOARD
  return (
    (row / strokeWidth[hatchPen] + col / strokeWidth[hatchPen]) % 2 ?
    primary :
    secondary
  );
#else
  return (
    (row + col) / strokeWidth[hatchPen] % 2 ?
    primary :
    secondary
  ); 
#endif
}


/*********************************************************
* Main                                                   *
*********************************************************/

int main(const int argc, const char* argv[])
{
  options.parse(argc, argv);

  if (options.verbose) { cerr << "> Preprocessing input..."; cerr.flush(); }

  // get sites
  int nrSites; cin >> nrSites;
  if (nrSites < 1) Complaint(userError) << "Too few sites";
  if (nrSites > MAXNRSITES) Complaint(userError) << "Too many sites";
  std::vector<Site> sites(nrSites);
  for (int i = 0; i < nrSites; ++i)
  {
    cin >> sites[i]; 
    sites[i].setId(i); // <- ugly solution to pass id to highway distance...
    options.distance->verifySite(sites[i]);
  }

  // get highways
  int nrHighways = 0;
  std::vector<Site> highways;
  if (options.useHighways) 
  {
    cin >> nrHighways;
    highways.resize(nrHighways);
    if (nrHighways < 1) Complaint(userError) << "Missing highways";
    if (nrHighways > MAXNRHIGHWAYS) Complaint(userError) << "Too many highways";
    for (int i = 0; i < nrHighways; ++i) cin >> highways[i];
    highwayMap = new HighwayMap(sites, highways);
  } 
  
  if (options.verbose) { cerr << "\r> Preparing canvas...   "; cerr.flush(); }

  Canvas canvas(options.width, options.width);
  if (options.verbose) { cerr << "\r> Preparing scan...     "; cerr.flush(); }

  // reserve space for regions and distances of last two line
  // (one line and current position would be enough, but would make the code
  // confusing)
  Owner* owners = new Owner[2 * options.width];
  Owner* owner[2]; 
  owner[0] = owners; 
  owner[1] = owners + options.width;
  
  DOUBLE* distances = new DOUBLE[2 * options.width * nrSites];
  DOUBLE** distance[2]; 
  distance[0] = new DOUBLE*[options.width]; 
  distance[1] = new DOUBLE*[options.width];
  for (int row = 0; row < 2; ++row)
    for (int col = 0; col < options.width; ++col)
    {
      distance[row][col] = distances;
      distances += nrSites;
    }

  // draw regions
  for (int row = 0; row < options.width; ++row)
  {
    if (options.verbose) 
      reportProgress(((DOUBLE) row)/options.width, "> Calculating regions");
    int cRow = row % 2;
    int pRow = 1 - cRow;
    DOUBLE y = (2.0 * row + 1.0) / options.width - 1.0;
    for (int col = 0; col < options.width; ++col)
    {
      Point p((2.0 * col + 1.0) / options.width - 1.0, y);
      for (int i = 0; i < nrSites; ++i) 
      {
        if (options.weighting == Options::divisive && sites[i].weight == 0)
        {
          distance[cRow][col][i] = INF;
          continue;
        }
        distance[cRow][col][i] = (*options.distance)(p, sites[i]);
        switch(options.weighting)
        {
          case Options::subtractive: 
            distance[cRow][col][i] -= sites[i].weight; 
            break;
          case Options::divisive:
            distance[cRow][col][i] /= sites[i].weight; 
            break;
        }
        if (!options.drawStandardContours || col == 0 || row == 0) continue;
        bool inner = (distance[cRow][col][i] <= options.contourDistance);
        if (
          (distance[cRow][col-1][i] <= options.contourDistance) != inner ||
          (distance[pRow][col  ][i] <= options.contourDistance) != inner ||
          (distance[pRow][col-1][i] <= options.contourDistance) != inner
        )
          canvas.draw(row, col, contourPen, RGB::black);
      }
      owner[cRow][col].select(distance[cRow][col], nrSites);
      int rep = owner[cRow][col].representative(row, col);
      canvas.shade(row, col, options.colours[rep % options.nrColours]);
      if (options.shadeInnerAreas && distance[cRow][col][rep] <= options.contourDistance)
        canvas.shade(row, col, RGB::grey);
      if (options.halfBrightDistance > 0)
        canvas.shade(row, col,
          RGB(pow(0.5, distance[cRow][col][rep] / options.halfBrightDistance)));
      if (options.contourInterval > 0 && col > 0 && row > 0)
      {
        int layer = distance[cRow][col][rep] / options.contourInterval;
        if (
          layer != (int) (distance[cRow][col-1][rep] / options.contourInterval) ||
          layer != (int) (distance[pRow][col  ][rep] / options.contourInterval) ||
          layer != (int) (distance[pRow][col-1][rep] / options.contourInterval) ||
          (
            (distance[cRow][col-1][rep] < 0) != (distance[cRow][col][rep] < 0)||
            (distance[pRow][col  ][rep] < 0) != (distance[cRow][col][rep] < 0)||
            (distance[pRow][col-1][rep] < 0) != (distance[cRow][col][rep] < 0)
          )
        )
        {
          canvas.shade(row,   col,   RGB::grey);
          canvas.shade(row,   col-1, RGB::grey);
          canvas.shade(row-1, col,   RGB::grey);
          canvas.shade(row-1, col-1, RGB::grey);
        }
      }
      if (!options.drawRegionBoundaries || col == 0 || row == 0) continue;
      if (
        owner[cRow][col-1] != owner[cRow][col] ||
        owner[pRow][col  ] != owner[cRow][col] ||
        owner[pRow][col-1] != owner[cRow][col]
      )
        canvas.draw(row, col, bisectorPen, RGB::black);
    }
  }
  if (options.verbose) reportProgress(1, "> Calculating regions");

  if (nrHighways > 0)
  {
    if (options.verbose) { cerr << "\r> Drawing highways...        "; cerr.flush(); }

    // draw the highways
    for (int i = 0; i < nrHighways; ++i)
      drawSite(canvas, highways[i], sitePen, RGB::black);
    for (int i = 0; i < nrHighways; ++i)
      drawSite(canvas, highways[i], bisectorPen, RGB::grey);
    // highwayMap->showNodes(canvas); // for debugging only
  }

  if (options.drawSites)
  {
    if (options.verbose) { cerr << "\r> Drawing sites...           "; cerr.flush(); }

    // draw the sites
    for (int i = 0; i < nrSites; ++i)
    {
      if (options.nrColours > 1)
      {
        drawSite(canvas, sites[i], siteOutlinePen, RGB::black);
        drawSite(canvas, sites[i], sitePen, options.colours[i % options.nrColours]);
      }
      else
        drawSite(canvas, sites[i], sitePen, RGB::black);
    }
  }

  // mark the origin
  if (options.markOrigin)
    for (
      int row = options.width/2 - 5 * strokeWidth[contourPen];
      row < options.width/2 + 5 * strokeWidth[contourPen];
      ++row)
    {
      canvas.draw(row, options.width/2, contourPen, RGB::black);     
      canvas.draw(options.width/2, row, contourPen, RGB::black);     
    }

  if (options.verbose) { cerr << "\r> Writing bitmap file...     "; cerr.flush(); }
  // output bitmap
  BitmapFile bmp(cout, options.width, options.width);
  for (int row = 0; row < options.width; ++row)
    for (int col = 0; col < options.width; ++col)
      bmp.writePixel(canvas(row,col));
  if (options.verbose) cerr << "\r> All done!              " << endl;
}
